import subprocess, shlex, os, time
from StringIO import StringIO
import whatap.util.file_util as file_util
from whatap.util.ip_util import IPUtil as ip_util
from whatap.io.data_outputx import DataOutputX
import whatap.util.process_util as process_util
import whatap.util.logging_util as logging_util
import whatap.util.thread_util as thread_util
import whatap.agent.conf.configure as conf
import psutil
import threading

import multiprocessing

cpucore = float(multiprocessing.cpu_count())


class ProcessPerfInfo:
    def __init__(self):
        self.ppid = None
        self.pid = None
        self.cpu = None
        self.memoryBytes = None
        self.memoryPercent = None
        self.readBps = 0.0
        self.writeBps = 0.0
        self.cmd1 = None
        self.cmd2 = None
        self.readIops = 0.0
        self.writeIops = 0.0
        self.user = None
        self.state = None
        self.createTime = None
        self.net = []
        self.file = []
        self.sharedMemory = 0

    def __str__(self):
        b = StringIO()
        b.write('ppid:')
        b.write(self.ppid)
        b.write('\npid')
        b.write(self.pid)
        b.write('\ncpu')
        b.write(self.cpu)
        b.write('\nmemory')
        b.write(self.memoryBytes)
        b.write('\nsharedMemory')
        b.write(self.sharedMemory)


        return b.getvalue()


class IpPort:
    def __init__(self, localip=0, localport=0, conn=0, ilocaladdr=0, inode=0):
        self.localip = localip
        self.localport = localport
        self.ilocaladdr = ilocaladdr
        self.iremoteaddr = 0
        self.conn = conn
        self.inode = inode

    def __str__(self):
        buf = StringIO()
        buf.write("localip=")
        buf.write(self.localip)
        buf.write("\nlocalport=")
        buf.write(self.localport)
        buf.write("\nilocaladdr=")
        buf.write(self.ilocaladdr)
        buf.write("\niremoteaddr=")
        buf.write(self.iremoteaddr)
        buf.write("\nconn=")
        buf.write(self.conn)
        buf.write("\ninode=")
        buf.write(self.inode)

        return buf.getvalue()


class FileInfo:
    def __init__(self, name=None, size=0):
        self.name = name
        self.size = size

def populateNetworkStatus(processPerfs):
    localips = []
    ipinfo = process_util.executeCmd("ip addr show")

    for line in ipinfo.split('\n'):
        words = line.split()

        if words and words[0] == 'inet':
            localips.append(words[1].split('/')[0])
    inodeportlookup = {}
    ipinodelookup = {}
    remoteconns = []
    for tcpfilepath in ("/proc/net/tcp", "/proc/net/tcp6"):
        tcpContent = file_util.readContent(tcpfilepath)
        if tcpContent:
            i = 0
            for line in tcpContent.split("\n"):
                i += 1
                if i == 1 or not line:
                    continue

                words = line.split()
                inode = int(words[9])
                localIPBytes = ip_util.parseHexString(words[1])
                if len(localIPBytes) != 6:
                    continue
                localIP = DataOutputX.toInt(localIPBytes[:4])
                localPort = DataOutputX.toShort(localIPBytes[4:])
                localIPPortInt = DataOutputX.toLong6(localIPBytes)
                remoteIPBytes = ip_util.parseHexString(words[2])
                if len(remoteIPBytes) != 6:
                    continue
                remoteIPInt = DataOutputX.toInt(remoteIPBytes[:4])
                remotePortInt = DataOutputX.toShort(remoteIPBytes[4:])
                ipport = IpPort(localip=localIP, localport=localPort, conn=0, ilocaladdr=localIPPortInt,
                                inode=int(inode))
                if remoteIPInt == 0 and remotePortInt == 0:
                    inodeportlookup[inode] = ipport
                    if localIPPortInt > 0:
                        ipinodelookup[localIPPortInt] = ipport
                        for ip in localips:
                            ipbytes = ip_util.toBytes(ip)
                            deviceIPPortBytes = StringIO()
                            deviceIPPortBytes.write(ipbytes)
                            deviceIPPortBytes.write(localIPBytes[4:])
                            deviceIPPortInt = DataOutputX.toLong6(deviceIPPortBytes.getvalue())
                            ipinodelookup[deviceIPPortInt] = ipport

                else:
                    remoteconns.append(ipport)

    for pipport in remoteconns:
        if pipport.ilocaladdr in ipinodelookup:
            ipinodelookup[pipport.ilocaladdr].conn += 1

    # print 'ipinodelookup'
    # from pprint import pprint
    # for k, v in ipinodelookup.items():
    #    print k,v

    for processPerf in processPerfs:
        ports = []
        filesinfo = []
        searchDir = os.path.join("/proc", str(processPerf.pid), "fd")

        if not os.path.exists(searchDir):
            continue
        for filename in os.listdir(searchDir):
            path = os.path.join(searchDir, filename)

            if os.path.islink(path):
                link = os.readlink(path)

                if link.startswith("socket:["):
                    inode = int(link[8:len(link) - 1])
                    # print 'inode', inode, inode in inodeportlookup
                    if inode in inodeportlookup:
                        ipport = inodeportlookup[inode]
                        ports.append(IpPort(localip=ipport.localip, localport=ipport.localport, conn=ipport.conn))

                elif len(link) > 0 and not link.startswith("/dev") and link[0] == '/':
                    try:
                        filestat = os.stat(link)
                        # print 'file ',processPerf.cmd1, link, filestat.st_size
                        if filestat and link.endswith(".log"):
                            fileinfo = FileInfo(name=link, size=filestat.st_size)
                            filesinfo.append(fileinfo)
                            # print processPerf.cmd1, link
                    except Exception, e:
                        pass
        processPerf.net = ports
        processPerf.file = filesinfo

class _ProcPerfCache:
    def __init__(self, cpu_times=None, proc_map=None):
        self.cpu_times = cpu_times
        self.proc_map = proc_map
        self.timestamp = time.time()


oldProcPerfCache = None

jiff = 100

def parseNetwork(conns):
    port_map = {}
    ports= []
    for c in conns:
        if c.family == 2 and c.type == 2 and c.status=='LISTEN':
            ipport = IpPort(localip=ip_util.strToInt(c.laddr.ip), localport=c.laddr.port, conn=0)
            port_map[c.laddr.port] = ipport
            ports.append(ipport)
    for c in conns:
        if c.family == 2 and c.type == 2 and c.status == 'ESTABLISHED' and c.laddr.port in port_map:
            port_map[c.laddr.port].conn += 1
    return ports

def parseOpenFile(files):
    flookup = {}
    fileusage =[]
    for f in files:
        if f.path and f.path.endswith('.log'):
            flookup[f.path] = 0
    for p in flookup.keys():
        filestat = os.stat(p)
        if filestat:
            fileusage.append(FileInfo(name=p, size=filestat.st_size))

    return fileusage



proc_shm_cache = {}
data_map_lock = threading.Lock()
update_condition = threading.Condition()
update_queue = []

def updateShmCache():
    while True:
        with update_condition:
            while not update_queue:
                update_condition.wait()

            proc = update_queue.pop(0)

        with data_map_lock:
            sharedMemory = 0
            try:
                if proc.pid not in proc_shm_cache or proc.create_time() != proc_shm_cache[proc.pid][1]:
                    for m in proc.memory_maps():
                        if m.path == '[shmid]':
                            sharedMemory += m.rss
                proc_shm_cache[proc.pid] = (sharedMemory, proc.create_time())
            except psutil.ZombieProcess, e:
                logging_util.debug("zombie process skip", e)
                continue
            except psutil.NoSuchProcess, e: 
                logging_util.debug("process cpu time error : ", e)
                continue


def setSharedMemory(proc):
    if data_map_lock.acquire(False):
        try:
            if proc.pid in proc_shm_cache and proc.create_time() == proc_shm_cache[proc.pid][1]:
                (sharedMemory, _) = proc_shm_cache[proc.pid]
                return sharedMemory
            else:
                with update_condition:
                    update_queue.append(proc)
                    update_condition.notify()
                return 0

        finally:
            data_map_lock.release()
    else:
        with update_condition:
            update_queue.append(proc)
            update_condition.notify()
        return 0


def getProcessPerfList():
    global proc_shm_cache
    proc_map = {}

    vmem = psutil.virtual_memory()
    totalMem = vmem.total
    proc_perfs = []

    cpu_times = psutil.cpu_times(percpu=False)
    cpu_jiff_diff = 0

    global oldProcPerfCache
    if oldProcPerfCache:
        all_cpu = time.time()
        last_all_cpu = oldProcPerfCache.timestamp
        cpu_jiff_diff = float(all_cpu - last_all_cpu) * cpucore

    for proc in psutil.process_iter():
        with proc.oneshot():
            try:
                proc_cpu_times = proc.cpu_times()
                proc_map[proc.pid] = (proc, proc_cpu_times)
                
                if not oldProcPerfCache:
                    continue
                else:
                    if proc.pid == 2 or proc.ppid() == 2 or proc.pid not in oldProcPerfCache.proc_map:
                        continue
                    p = ProcessPerfInfo()
                    p.pid = proc.pid
                    p.ppid = proc.ppid()
                    p.user = proc.username()
                    p.memoryBytes = proc.memory_info().rss
                    p.state = proc.status()

                    _, old_cpu_times = oldProcPerfCache.proc_map[proc.pid]

                    if cpu_jiff_diff != 0:
                        now = proc_cpu_times.user + proc_cpu_times.system
                        old = old_cpu_times.user + old_cpu_times.system
                        p.cpu = float(100) * float(now - old) / cpu_jiff_diff
                    p.memoryPercent = 100.0 * p.memoryBytes / totalMem
                    p.cmd1 = proc.exe()
                    p.cmd2 = ' '.join(proc.cmdline())
                    p.createTime = int(proc.create_time())
                    p.readBps = float(0)
                    p.writeBps = float(0)
                    p.readIops = float(0)
                    p.writeIops = float(0)

                    if conf.GetConfig().serverProcessConnectCheck:
                        p.net = parseNetwork(proc.connections())
                    if conf.GetConfig().serverProcessFDCheck:
                        p.file = parseOpenFile(proc.open_files())

                    proc_perfs.append(p)
            except psutil.ZombieProcess, e:
                continue
            except psutil.NoSuchProcess, e:
                continue

    procPerfCache = _ProcPerfCache(cpu_times=cpu_times, proc_map=proc_map)
    oldProcPerfCache = procPerfCache
    return proc_perfs

init=False

def GetProc():
    global init
    if init:
        thread_util.async(updateShmCache)
        init=False

    logging_util.debug("Get Proc Start")
    procPerfs = getProcessPerfList()
    if procPerfs == None:
        logging_util.debug("Get Proc End : None")
    else:
        logging_util.debug("Get Proc End : ", len(procPerfs))

    return procPerfs


def test():
    while True:
        procs = GetProc()
        print
        procs
        if not procs:
            time.sleep(5)
            continue
        time.sleep(5)


if __name__ == '__main__':
    test()
